import asyncio
import contextlib
from typing import Awaitable, Any
import atexit

__version__ = "0.1.0"

# One persistent event loop
_loop = None
_nest_asyncio_applied = False

def loop() -> asyncio.AbstractEventLoop:

    global _loop, _nest_asyncio_applied

    # Try to get the currently running loop (e.g., in Jupyter)
    try:
        running_loop = asyncio.get_running_loop()
        # We found a running loop (Jupyter case)
        if not _nest_asyncio_applied:
            try:
                import nest_asyncio
                nest_asyncio.apply(running_loop)
                _nest_asyncio_applied = True
            except ImportError:
                pass
        return running_loop
    except RuntimeError:
        # No running loop, create our own
        pass

    if _loop is None or _loop.is_closed():
        _loop = asyncio.new_event_loop()
        asyncio.set_event_loop(_loop)

        # Apply nest_asyncio to our new loop
        if not _nest_asyncio_applied:
            try:
                import nest_asyncio
                nest_asyncio.apply(_loop)
                _nest_asyncio_applied = True
            except ImportError:
                pass

    return _loop
loop()  # Make sure to initialize `_loop`

def _reap_done_tasks(evloop: asyncio.AbstractEventLoop) -> None:
    """Reap exceptions from tasks that are already DONE on this loop.
    Does not cancel or otherwise touch pending tasks.
    """
    if evloop is None or evloop.is_closed():
        return

    try:
        # Snapshot; tasks may change during iteration
        for t in tuple(asyncio.all_tasks(evloop)):
            if not t.done():
                continue
            # If the task is cancelled, .result() raises CancelledError — suppress that.
            # If the task failed, .result() raises its exception — suppress to just mark it retrieved.
            with contextlib.suppress(asyncio.CancelledError, Exception):
                t.result()
    except (RuntimeError, AttributeError):
        # During shutdown, asyncio may be partially cleaned up
        pass


def arun(coro: Awaitable[Any]) -> Any:

    evloop = loop()

    try:
        return evloop.run_until_complete(coro)
    finally:
        # Clean up completed/cancelled tasks so residual CancelledError
        # doesn't leak to next run
        _reap_done_tasks(evloop)




from .float_signal import FloatSignalContainer
from .types import EpicsConfigR

PYAMLCLASS : str = "EpicsR"

class ConfigModel(EpicsConfigR):
    unit: str = ""

class EpicsR(FloatSignalContainer):
    def __init__(self, cfg: ConfigModel, is_array=False):
        super().__init__(cfg,is_array)
    def get_cs(self) -> str:
        return "epics"        


from ophyd_async.epics.signal import epics_signal_r, epics_signal_w, epics_signal_rw
from ophyd_async.core import Array1D
import numpy

from .container import OAReadback as Readback
from .container import OASetpoint as Setpoint
from .types import (
    ControlSysConfig,
    EpicsConfigR,
    EpicsConfigRW,
    EpicsConfigW,
)


def get_SP_RB(cfg: ControlSysConfig,is_array:bool) -> tuple[Setpoint | None, Readback | None]:
    setpoint: Setpoint | None = None
    readback: Readback | None = None

    assert isinstance(cfg, (EpicsConfigRW, EpicsConfigR, EpicsConfigW))

    if isinstance(cfg, EpicsConfigR):
        r_sig = epics_signal_r(
            datatype=float if not is_array else Array1D[numpy.float64],
            read_pv=cfg.read_pvname,
            name="",
            timeout = cfg.timeout_ms / 1000.,
        )
        readback = Readback(r_sig)
        setpoint = None

    if isinstance(cfg, EpicsConfigW):
        w_sig = epics_signal_w(
            datatype=float if not is_array else Array1D[numpy.float64],
            write_pv=cfg.write_pvname,
            name="",
            timeout = cfg.timeout_ms / 1000.,
        )
        readback = None
        setpoint = Setpoint(w_sig)

    if isinstance(cfg, EpicsConfigRW):
        w_sig = epics_signal_rw(
            datatype=float if not is_array else Array1D[numpy.float64],
            read_pv=cfg.read_pvname,
            write_pv=cfg.write_pvname,
            name="",
            timeout = cfg.timeout_ms / 1000.,
        )
        readback = Readback(w_sig)
        setpoint = Setpoint(w_sig)


    return setpoint, readback



from ophyd_async.tango.core import tango_signal_r, tango_signal_rw
from ophyd_async.core import Array1D
import numpy

from .container import OAReadback as Readback
from .container import OASetpoint as Setpoint
from .types import (
    ControlSysConfig,
    TangoConfigR,
    TangoConfigRW,
)


def get_SP_RB(cfg: ControlSysConfig,is_array:bool) -> tuple[Setpoint | None, Readback | None]:
    setpoint: Setpoint | None = None
    readback: Readback | None = None

    assert isinstance(cfg, (TangoConfigRW, TangoConfigR))

    if isinstance(cfg, (TangoConfigR)):
        r_sig = tango_signal_r(
            datatype=float if not is_array else Array1D[numpy.float64],
            read_trl=cfg.attribute,
            timeout=cfg.timeout_ms,
        )
        readback = Readback(r_sig)
        setpoint = None

    elif isinstance(cfg, (TangoConfigRW)):
        rw_sig = tango_signal_rw(
            datatype=float if not is_array else Array1D[numpy.float64],
            read_trl=cfg.attribute,
            write_trl=cfg.attribute,
            timeout=cfg.timeout_ms,
        )
        readback = Readback(rw_sig)
        setpoint = Setpoint(rw_sig)

    return setpoint, readback



from .float_signal import FloatSignalContainer
from .types import TangoConfigR

PYAMLCLASS : str = "TangoR"

class ConfigModel(TangoConfigR):
    unit: str = ""

class TangoR(FloatSignalContainer):
    def __init__(self, cfg: ConfigModel, is_array=False):
        super().__init__(cfg,is_array)
    def get_cs(self) -> str:
        return "tango"        


import os
import logging
import copy

from pyaml.control.controlsystem import ControlSystem
from pydantic import BaseModel
from pyaml.common.exception import PyAMLException

PYAMLCLASS : str = "OphydAsyncControlSystem"

logger = logging.getLogger(__name__)

from .types import (
    EpicsConfigR,
    EpicsConfigW,
    EpicsConfigRW,
    TangoConfigR,
    TangoConfigRW,
)
from .signal import OASignal
from .epicsR import EpicsR
from .epicsW import EpicsW
from .epicsRW import EpicsRW
from .tangoR import TangoR
from .tangoRW import TangoRW

class ConfigModel(BaseModel):

    """
    Configuration model for an OA Control System.

    Attributes
    ----------
    name : str
        Name of the control system.
    prefix : str
        Prefix added to the PV or attribute name. It can be a 
        for instance, TANGO_HOST, or a PV prefix.
    debug_level : int
        Debug verbosity level.
    scalar_aggregator : str
        Aggregator module for scalar values. If none specified, writings and 
        readings of sclar value are serialized. 
    vector_aggregator : str
        Aggregator module for vecrors. If none specified, writings and readings
        of vector are serialized,
    """

    name: str
    prefix: str = ""
    debug_level: str=None
    scalar_aggregator: str | None = "pyaml_cs_oa.scalar_aggregator"
    vector_aggregator: str | None = None


class OphydAsyncControlSystem(ControlSystem):
    """A generic control system using ophyd_async backend."""

    def __init__(self, cfg: ConfigModel):
        super().__init__()
        self._cfg = cfg
        self._devices = {} # Dict containing all attached DeviceAccess

        if self._cfg.debug_level:
          log_level = getattr(logging, self._cfg.debug_level, logging.WARNING)
          logger.parent.setLevel(log_level)
          logger.setLevel(log_level)

        logger.log(logging.WARNING, f"OA control system binding for PyAML initialized with name '{self._cfg.name}'"
                                 f" and prefix='{self._cfg.prefix}'")
        
    def attach(self, devs: list[OASignal]) -> list[OASignal]:
        return self._attach(devs,False)

    def attach_array(self, devs: list[OASignal]) -> list[OASignal]:
        return self._attach(devs,True)

    def _attach(self, devs: list[OASignal],is_array:bool) -> list[OASignal]:
        # Concatenate the prefix
        newDevs = []
        for d in devs:            
            if d is not None:

                sig_cfg = d._cfg
                sig_cfg_cls = sig_cfg.__class__
                
                if isinstance(d._cfg,EpicsConfigR):
                    key = self._cfg.prefix + d._cfg.read_pvname
                    sig_cls = EpicsR
                    config = dict(read_pvname=key,timeout_ms=d._cfg.timeout_ms)
                elif isinstance(d._cfg,EpicsConfigW):
                    key = self._cfg.prefix + d._cfg.write_pvname
                    sig_cls = EpicsW
                    config = dict(write_pvname=key,timeout_ms=d._cfg.timeout_ms)
                elif isinstance(d._cfg,EpicsConfigRW):
                    key = self._cfg.prefix + d._cfg.read_pvname + d._cfg.write_pvname
                    sig_cls = EpicsRW
                    config = dict(read_pvname=self._cfg.prefix + d._cfg.read_pvname, write_pvname=self._cfg.prefix + d._cfg.write_pvname,timeout_ms=d._cfg.timeout_ms)
                elif isinstance(d._cfg,TangoConfigR):
                    key = self._cfg.prefix + d._cfg.attribute
                    sig_cls = TangoR
                    config = dict(attribute=key,timeout_ms=d._cfg.timeout_ms)
                elif isinstance(d._cfg,TangoConfigRW):
                    key = self._cfg.prefix + d._cfg.attribute
                    sig_cls = TangoRW
                    config = dict(attribute=key,timeout_ms=d._cfg.timeout_ms)
                else:
                    raise PyAMLException(f"OphydAsyncControlSystem: Unsupported type {type(sig_cfg)}")

                if key not in self._devices:
                    nr = sig_cls(sig_cfg_cls(**config),is_array)
                    nr.build()
                    self._devices[key] = nr

                newDevs.append(self._devices[key])
            else:
                newDevs.append(None)
        return newDevs

    def name(self) -> str:
        """
        Return the name of the control system.

        Returns
        -------
        str
            Name of the control system.
        """
        return self._cfg.name
    
    def scalar_aggregator(self) -> str | None:
        """
        Returns the module name used for handling aggregator of DeviceAccess

        Returns
        -------
        str
            Aggregator module name
        """
        return self._cfg.scalar_aggregator

    def vector_aggregator(self) -> str | None:
        """
        Returns the module name used for handling aggregator of DeviceVectorAccess
        
        Returns
        -------
        str
            Aggregator module name
        """
        return self._cfg.vector_aggregator

    def __repr__(self):
       return repr(self._cfg).replace("ConfigModel",self.__class__.__name__)


from pyaml.control.deviceaccess import DeviceAccess

from .types import (
    ControlSysConfig,
    EpicsConfigR,
    EpicsConfigRW,
    EpicsConfigW,
    TangoConfigR,
    TangoConfigRW,
)

class OASignal(DeviceAccess):
    """
    Class that implements a PyAML Signal using ophyd_async Signals.
    """

    def __init__(self, cfg: ControlSysConfig,is_array:bool):
        self._cfg = cfg
        self.is_array = is_array

    def build(self):

        self._readable: bool = isinstance(
            self._cfg, (EpicsConfigR, TangoConfigR)
        )
        self._writable: bool = isinstance(
            self._cfg, (EpicsConfigRW, EpicsConfigW, TangoConfigRW)
        )

        cs_name = self.get_cs()
        if cs_name == "tango":
            from .tango import get_SP_RB
        elif cs_name == "epics":
            from .epics import get_SP_RB
        else:
            raise ValueError(f"Unsupported cs_name: {cs_name}")

        self.SP, self.RB = get_SP_RB(self._cfg,self.is_array)
        if self.SP:
            self.SP.__peer__ = self
        if self.RB:
            self.RB.__peer__ = self

    def get_cs(self) -> str:
        raise Exception("get_cs() not implemented")

    def name(self) -> str:
        """
        Return the name of the signal.
        """
        return self._signal.name

    def measure_name(self) -> str:
        """
        Return the short attribute name (last component).

        Returns
        -------
        str
            The attribute name (e.g., 'current').
        """

        # TODO override measure name in sub classes
        if isinstance(self._cfg, (EpicsConfigR, EpicsConfigRW)):
            return self._cfg.read_pvname
        elif isinstance(self._cfg, (TangoConfigR, TangoConfigRW)):
            return self._cfg.attribute
        else:
            raise ValueError(
                f"Unsupported control system config type: {type(self._cfg)!r}"
            )

    def unit(self) -> str:
        """
        Return the unit of the attribute.

        Returns
        -------
        str
            The unit string.
        """
        return self._cfg.unit

    def __repr__(self):
       cfg_str = repr(self._cfg)
       # Replace the pydantic config class by the class itself
       idx = cfg_str.find("(")
       return f"{self.__class__.__name__}{cfg_str[idx:]}"



import numpy as np
from numpy import typing as npt
from pyaml.control.deviceaccess import DeviceAccess
from pydantic import BaseModel
from pyaml.control.deviceaccesslist import DeviceAccessList
from pyaml import PyAMLException

from .float_signal import FloatSignalContainer
from . import arun

import asyncio

PYAMLCLASS : str = "OAScalarAggregator"


class ConfigModel(BaseModel):
    pass

class OAScalarAggregator(DeviceAccessList):

    def __init__(self, cfg:ConfigModel=None):
        super().__init__()

    def add_devices(self, devices: DeviceAccess | list[DeviceAccess]):
        if isinstance(devices, list):
            if any([not isinstance(device, FloatSignalContainer) for device in devices]):
                raise PyAMLException("All devices must be instances of FloatSignalContainer.")
            super().extend(devices)
        else:
            if not isinstance(devices, FloatSignalContainer):
                raise PyAMLException("Device must be an instance of FloatSignalContainer.")
            super().append(devices)

    def get_devices(self) -> DeviceAccess | list[DeviceAccess]:
        if len(self)==1:
            return self[0]
        else:
            return self


    def set(self, value: npt.NDArray[np.float64]):
        
        if len(value)!=len(self):
            raise PyAMLException(f"Size of value ({len(value)} do not match the number of managed devices ({len(self)})")
        
        d: FloatSignalContainer
        requests = [] # list of status to await
        for idx,d in enumerate(self):
            requests.append( d.SP._complete_set(value[idx]) )
        arun(asyncio.gather(*requests))

    def set_and_wait(self, value: npt.NDArray[np.float64]):
        raise NotImplemented("Not implemented yet.")

    def get(self) -> npt.NDArray[np.float64]:

        d: FloatSignalContainer
        requests = [] # list of status to await
        for d in self:
            if d._writable:
                requests.append( d.SP.async_get() )
            else:
                requests.append( d.RB.async_get() )
        values = arun(asyncio.gather(*requests))
        return np.array(values)

    def readback(self) -> np.array:

        d: FloatSignalContainer
        requests = [] # list of status to await
        for d in self:
            requests.append( d.RB.async_get() )
        values = arun(asyncio.gather(*requests))
        return np.array(values)

    def __repr__(self):
       return repr(self._cfg).replace("ConfigModel",self.__class__.__name__)


import asyncio
import inspect
from collections.abc import Awaitable, Callable
from typing import TypeVar
from .signal import OASignal

from ophyd_async.core import (
    SignalDatatypeT,
    SignalR,
    SignalW,
    set_and_wait_for_other_value,
)

T = TypeVar("T")

from . import arun

def _looks_disconnected(exc: BaseException) -> bool:
    # Keep it generic: ophyd-async wraps cancellations in TimeoutError;
    # tango/epics transports often raise CancelledError or "NotConnected" types.
    return isinstance(exc, (asyncio.CancelledError, TimeoutError))


async def _recover_once(
    run: Callable[[], Awaitable[T]],
    reconnect: Callable[[], Awaitable[None]],
    peer: OASignal,
) -> T:
    try:
        return await run()
    except BaseException as exc:
        if not _looks_disconnected(exc):
            raise
        # Attempt reconnect of the same Signal first
        try:
            await reconnect()
            return await run()
        except BaseException:
            # If that fails and we have a way to rebuild, do so and try one more time
            if peer is not None:
                maybe_awaitable = peer.build()
                if inspect.isawaitable(maybe_awaitable):
                    await maybe_awaitable
                await reconnect()
                return await run()
            raise


class OAReadback():
    """A readback object."""

    def __init__(self, r_signal: SignalR[SignalDatatypeT]):
        self._r_sig = r_signal

    async def _run_get(self) -> SignalDatatypeT:
        await self._r_sig.connect()
        backend = self._r_sig._connector.backend
        return await backend.get_value()

    async def async_get(self) -> SignalDatatypeT:
        return await _recover_once(
            self._run_get,
            self._r_sig.connect,
            getattr(self._r_sig, "__peer__", None),
        )

    async def _run_read(self) -> SignalDatatypeT:
        await self._r_sig.connect()
        backend = self._r_sig._connector.backend
        return await backend.get_reading()

    async def async_read(self) -> SignalDatatypeT:
        return await _recover_once(
            self._run_read,
            self._r_sig.connect,
            getattr(self._r_sig, "__peer__", None),
        )

    def get(self) -> SignalDatatypeT:
        """Synchronous wrapper around `async_get()`."""
        return arun(self.async_get())

class OASetpoint():
    def __init__(
        self,
        w_signal: SignalW[SignalDatatypeT],
        r_signal: SignalR[SignalDatatypeT] | None = None,
    ):
        self._w_sig = w_signal
        self._r_sig = r_signal # used only for `set_and_wait()`
        self._has_r_sig = (r_signal is not None)

    async def _run_get(self) -> SignalDatatypeT:
        await self._w_sig.connect()
        backend = self._w_sig._connector.backend
        return await backend.get_setpoint()

    async def async_get(self) -> SignalDatatypeT:
        return await _recover_once(
            self._run_get,
            self._w_sig.connect,
            getattr(self._w_sig, "__peer__", None),
        )

    async def _run_read(self) -> SignalDatatypeT:
        await self._w_sig.connect()
        backend = self._w_sig._connector.backend
        return await backend.get_reading()

    async def async_read(self) -> SignalDatatypeT:
        return await _recover_once(
            self._run_read,
            self._w_sig.connect,
            getattr(self._w_sig, "__peer__", None),
        )

    async def _run_set(self, value):
        await self._w_sig.connect()
        status = self._w_sig.set(value)
        return status

    async def async_set(self, value):
        return await _recover_once(
            lambda: self._run_set(value),
            self._w_sig.connect,
            getattr(self._w_sig, "__peer__", None),
        )

    async def _reconnect_both(self) -> None:
        await asyncio.gather(self._w_sig.connect(), self._r_sig.connect())

    async def _rebuild_both(self) -> None:
        w_rebuild = getattr(self._w_sig, "__peer__", None)
        r_rebuild = getattr(self._r_sig, "__peer__", None)
        if w_rebuild is not None:
            w_rebuild()
        if r_rebuild is not None:
            r_rebuild()

    async def _run_set_and_wait(self, value) -> None:
        if not self._has_r_sig:
            raise RuntimeError(
                "Cannot use set_and_wait() without a matching readback signal."
            )
        await self._reconnect_both()
        await set_and_wait_for_other_value(self._w_sig, value, self._r_sig, value)

    async def async_set_and_wait(self, value) -> None:
        return await _recover_once(
            lambda: self._run_set_and_wait(value),
            self._reconnect_both,
            self._rebuild_both,
        )

    async def _complete_set(self, value):
            status = await self.async_set(value)
            await status  # Wait for completion before returning
            return status

    def set(self, value):
        """Synchronous wrapper around `async_set()`."""
        return arun(self._complete_set(value))

    def get(self) -> SignalDatatypeT:
        """Synchronous wrapper around `async_get()`."""
        return arun(self.async_get())

    def set_and_wait(self, value) -> None:
        """Synchronous wrapper around `async_set_and_wait()`."""
        return arun(self.async_set_and_wait(value))


from pydantic import BaseModel


class EpicsConfigR(BaseModel):
    read_pvname: str
    timeout_ms: int = 3000

class EpicsConfigW(BaseModel):
    write_pvname: str
    timeout_ms: int = 3000


class EpicsConfigRW(BaseModel):
    read_pvname: str
    write_pvname: str
    timeout_ms: int = 3000


class TangoConfigR(BaseModel):
    attribute: str
    timeout_ms: int = 3000


class TangoConfigRW(BaseModel):
    attribute: str
    timeout_ms: int = 3000


ControlSysConfig = (
    EpicsConfigR
    | EpicsConfigW
    | EpicsConfigRW
    | TangoConfigR
    | TangoConfigRW
)



from .signal import OASignal
from .types import ControlSysConfig

class FloatSignalContainer(OASignal):
    """
    Class that implements a PyAML Float/FloatArray Signal using ophyd_async Signals.
    """

    def __init__(self, cfg: ControlSysConfig,is_array:bool):
        super().__init__(cfg,is_array)

    def get(self):
        """
        Get the last written value(s) of the attribute.

        Returns
        -------
        float | list[float]
            The last written value(s).

        """
        if self._writable:
            return self.SP.get()
        else:
            return self.RB.get()

    def readback(self):
        """
        Return the readback value(s) with metadata.

        Returns
        -------
        Value | list[Value]
            The readback value(s) including quality and timestamp.

        """
        return self.RB.get()

    def set(self, value):
        """
        Write a value asynchronously to the device.

        Parameters
        ----------
        value : float | list[float]
            Value(s) to write to the attribute.

        """
        return self.SP.set(value)

    def set_and_wait(self, value):
        """
        Write a value(s) synchronously to the device.

        Parameters
        ----------
        value : float | list[float]
            Value to write to the attribute.

        """
        self.SP.set_and_wait(value)



from pyaml.accelerator import Accelerator
from pyaml.magnet.magnet import Magnet
from pyaml.common.constants import ACTION_RESTORE

import numpy as np
import time


def tune_callback(step: int, action: int, m: Magnet, dtune: np.array):
    if action == ACTION_RESTORE:
        # On action restore, the measured delta tune is passed as argument
        print(f"Tune response: #{step} {m.get_name()} {dtune}")
    return True

sr = Accelerator.load("bessy2tune-KL.yaml")

#print(sr.live.get_magnets("QForTune").strengths.get())
#print(sr.design.get_magnets("QForTune").strengths.get())
#quit()

#sr.design.get_lattice().disable_6d()
#tune_adjust = sr.design.tune
#tune_adjust.response.measure(callback=tune_callback)
#tune_adjust.response.save_json("tunemat-bessy.json")

sr.live.tune.response.load_json("tunemat-bessy.json")
print(sr.live.tune.readback())
sr.live.tune.set([0.83, 0.84],iter=2,wait_time=3)
time.sleep(3)
print(sr.live.tune.readback())




from pyaml.accelerator import Accelerator
from pyaml.magnet.magnet import Magnet
from pyaml.common.constants import ACTION_RESTORE

import numpy as np
import time

def tune_callback(step: int, action: int, m: Magnet, dtune: np.array):
    if action == ACTION_RESTORE:
        # On action restore, the delta tune is passed as argument
        print(f"Tune response: #{step} {m.get_name()} {dtune}")
    return True

sr = Accelerator.load("EBSTune-ophyd.yaml",use_fast_loader=False)

#sr.design.get_lattice().disable_6d()
#tune_adjust = sr.design.get_tune_tuning("TUNE")
#tune_adjust.response.measure(callback=tune_callback)
#tune_adjust.response.save_json("tunemat.json")

sr.live.get_tune_tuning("TUNE").response.load_json("tunemat.json")
print(sr.live.get_tune_tuning("TUNE").readback())
sr.live.get_tune_tuning("TUNE").set([0.17, 0.32])
time.sleep(10)
print(sr.live.get_tune_tuning("TUNE").readback())



import sys
import yaml
import at


if len(sys.argv) != 3:
    print(f"Convert to KL convrention, basically multiply the calibration factor by the magnet length")
    print(f"Usage: {sys.argv[0]} filename.yaml lattice.mat")
    quit()

ring = at.load_lattice(sys.argv[2])


def get_length(names:str):
    name_list = []
    l = 0
    # Support only list() syntax
    if names.startswith("list("):
        try:
            name_list = names[5:-1].rsplit(",")
        except Exception as err:
            strErr = f"{names}: Invalid lattice_names syntax :{str(err)}"
            print(strErr)
            quit()
    else:
        name_list = [names]

    for n in name_list:
        elts = ring.get_elements(n)
        for e in elts:
            l += e.Length

    return l


def get_names(d) -> str:
    if "lattice_names" in d:
        return d["lattice_names"]
    else:
        return d["name"]


def convert(d,path:str):

    if isinstance(d, list):
        # list can be a list of objects or a list of native types
        l = []
        for _index, e in enumerate(d):
            if isinstance(e, dict) or isinstance(e, list):
                convert(e,f"{path}/{get_names(e)}")

    elif isinstance(d, dict):
        for key, value in d.items():
            if isinstance(value, dict) or isinstance(value, list):
                convert(value,f"{path}/{str(key)}")

        if d["type"]=="pyaml.magnet.linear_model":
            names = path.split("/")[-2] # Get elements names
            length = get_length(names) # Get total length
            d["calibration_factor"] *= length


with open(sys.argv[1]) as stream:
    try:
        d = yaml.safe_load(stream)
        convert(d,"")        
        yaml.dump(d,sys.stdout,sort_keys=False)
    except yaml.YAMLError as exc:
        print(f"Failed to open {sys.argv[1]}: {str(exc)}")


